#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <pthread.h>
#include <string.h>
#include <stddef.h>
#include <fcntl.h>
#include <errno.h>
#include <netdb.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <fcntl.h>

#define MAX_TAMPON		1024
#define MAX_CLIENT 		20
#define MAX_UDP_MESSAGE 	100
#define MAX_CONNEXIONS		30
#define MAX_MUTX 30
#define UDP_PORT 4848

int UDPsock, ServerLoop=1;
unsigned char nbThread=0;
pthread_mutex_t mutexes[MAX_MUTX];

typedef struct {
    void (*callback)(void *);
    void *arg;
}threadStruct;

void P(int sem){
	pthread_mutex_lock(mutexes+sem);
}

void V(int sem){
	pthread_mutex_unlock(mutexes+sem);//available for next operation
}

void *clientThread(void *arg){
	int status;
	threadStruct *thrStr;
   	thrStr=(threadStruct *)arg;
    	thrStr->callback(thrStr->arg);
    	free(arg);
	P(0);
	nbThread--;
	V(0);
	pthread_exit((void *)&status);
}


int lanceThread(void (*function)(void *), void *arg, int taille){
	pthread_t pthr;
	pthread_attr_t attr;

	threadStruct *thrStr = (threadStruct *)malloc(sizeof(threadStruct));
	if(thrStr==NULL) {perror("struct malloc");exit(EXIT_FAILURE);}
	thrStr->callback=function;
	if(taille>0){
    		thrStr->arg=(threadStruct *)malloc(taille);
		if(thrStr->arg==NULL) {perror("arg malloc");exit(EXIT_FAILURE);}
		
	}
	thrStr->arg=arg;
	
	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
	P(0);
	nbThread++;
	V(0);
	int err = pthread_create(&pthr, &attr, clientThread, (void *)thrStr);
	if(err){perror("pthread_create");exit(EXIT_FAILURE);}
	return pthr;

}
void writeToFile(unsigned char boardID, unsigned char dbID, int value){
	char buffer[4];
	char ch[3]= {'\0', '\0', '\0'};
	char fileName[90]={"/var/www/troisiemesexe.lol/graph/csv/temperature/"};
	ch[0] = 'A' + (dbID - 1);
	ch[1] = '/';
	strcat(fileName, ch);
	strcat(fileName, "capteur");
	snprintf(buffer, sizeof(buffer), "%03d", boardID);
	strcat(fileName, buffer);
	strcat(fileName, ".tsv");
	printf("Filename : %s\n", fileName);
	time_t t = time(NULL);
	struct tm mytm = *localtime(&t);


	FILE *fp=fopen(fileName, "r+");
	if(fp==NULL){ 
		fp=fopen(fileName, "w");
		if(fp==NULL){ perror("fopen"); exit(EXIT_FAILURE); }
		fprintf(fp, "date\tvalues\n");	
		FILE *fp2=fopen("/var/www/troisiemesexe.lol/graph/csv/sensors.txt", "a+");
		if(fp2==NULL){ perror("fopen: /var/www/troisiemesexe.lol/graph/csv/sensors.txt"); exit(EXIT_FAILURE); }
		fprintf(fp2, "%03d\n", boardID);
		fclose(fp2);
	}
	fseek(fp, 0, SEEK_END);
	fprintf(fp, "%d-%02d-%02dT%02d:%02d:%02d\t", mytm.tm_year + 1900, mytm.tm_mon + 1, mytm.tm_mday, mytm.tm_hour, mytm.tm_min, mytm.tm_sec);
	fprintf(fp, "%d\n", value);
	fclose(fp);

}
void processUDPMessage(void *arg){
	int value=0;
	char buffer[4]={'\0', '\0', '\0', '\0'};
	unsigned char *chaine=((unsigned char *) arg);
	strncpy(buffer, (const char *)chaine+6, 3);
	value= atoi(buffer);
	strncpy(buffer, (const char *)chaine+1, 3);
	int board=atoi(buffer);
	
	buffer[0]='\0'; buffer[1]='\0'; buffer[2]='\0'; buffer[3]='\0';
	strncpy(buffer, (const char *)chaine+4, 1);	
	int db=atoi(buffer);
	printf("Board : %d DB: %d Msg : %s Value: %d\n", board, db, chaine, value);
	if((board >0 && board <=100) && (db>0 && db<=8))
		writeToFile(board, db, value);
	
}

void gestionMessage(unsigned char *msg, int taille){
	lanceThread((void *)processUDPMessage, msg, taille);
}




int initialisationSocketUDP(char *service){
	struct addrinfo precisions,*resultat;
	int statut;
	int s;

	/* Construction de la structure adresse */
	memset(&precisions,0,sizeof precisions);
	precisions.ai_family=AF_UNSPEC;
	precisions.ai_socktype=SOCK_DGRAM;
	precisions.ai_flags=AI_PASSIVE;
	statut=getaddrinfo(NULL,service,&precisions,&resultat);
	if(statut<0){ perror("initialisationSocketUDP.getaddrinfo"); exit(EXIT_FAILURE); }

	/* Creation d'une socket */
	s=socket(resultat->ai_family,resultat->ai_socktype,resultat->ai_protocol);
	if(s<0){ perror("initialisationSocketUDP.socket"); exit(EXIT_FAILURE); }

	/* Options utiles */
	int vrai=1;
	if(setsockopt(s,SOL_SOCKET,SO_REUSEADDR,&vrai,sizeof(vrai))<0){
	  perror("initialisationServeurUDPgenerique.setsockopt (REUSEADDR)");
	  exit(-1);
	}

	/* Specification de l'adresse de la socket */
	statut=bind(s,resultat->ai_addr,resultat->ai_addrlen);
	if(statut<0) {perror("initialisationServeurUDP.bind"); exit(-1);}

	/* Liberation de la structure d'informations */
	freeaddrinfo(resultat);

	return s;
}



void serveurMessages(short int port, void(*traitement)(unsigned char *, int)){//UDP serveur

	char service[5];
	sprintf(service, "%d", port);
	printf("Port UDP : %d\n", port);
	UDPsock=initialisationSocketUDP(service);
	while(ServerLoop){
		struct sockaddr_storage adresse;
		socklen_t taille=sizeof(adresse);
		unsigned char message[MAX_UDP_MESSAGE];
		int nboctets=recvfrom(UDPsock,message,MAX_UDP_MESSAGE,0,(struct sockaddr *)&(adresse),&taille);
		if(nboctets<0) exit(EXIT_FAILURE);			
		traitement(message, nboctets);	
	}

	close(UDPsock);
}







int main(int argc, char **argv){
	int i;
	for (i = 0; i < MAX_MUTX; i++)//init mutex
        pthread_mutex_init(&mutexes[i], NULL);
	serveurMessages(UDP_PORT, gestionMessage);
	return 0;
}
